##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Peter Eastman, Robert McGibbon
# Contributors: Kyle A. Beauchamp, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and Peter Eastman. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################

from __future__ import annotations

import itertools
import os
import warnings
import xml.etree.ElementTree as etree
from collections import namedtuple
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import TYPE_CHECKING, cast

import numpy as np
from numpy.typing import NDArray

from mdtraj.core import element as elem
from mdtraj.core.residue_names import (
    _AMINO_ACID_CODES,
    _NUCLEIC_ACID_CODES,
    _NUCLEIC_RESIDUES,
    _PROTEIN_RESIDUES,
    _WATER_RESIDUES,
)
from mdtraj.core.selection import parse_selection
from mdtraj.utils import ensure_type, ilen, import_
from mdtraj.utils.singleton import Singleton

if TYPE_CHECKING:
    import networkx as nx
    import openmm  # type: ignore
    import pandas as pd

    import mdtraj as md


def _topology_from_subset(topology: Topology, atom_indices: list[int]) -> Topology:
    """Create a new topology that only contains the supplied indices

    Note
    ----
    This really should be a copy constructor (class method) on Topology.
    It used to work on OpenMM topologies, but we've diverged where that no
    longer works.

    Parameters
    ----------
    topology : mdtraj.Topology
        The base topology
    atom_indices : array_like, dtype=int
        The indices of the atoms to keep
    """
    np_indices = np.array(atom_indices)
    if len(np_indices) > 1:
        if (np_indices[1:] - np_indices[:-1]).min() < 1:
            warnings.warn("atom_indices are not monotonically increasing")
        if len(np.unique(np_indices)) < len(np_indices):
            warnings.warn("atom_indices are not unique")

    newTopology = Topology()
    old_atom_to_new_atom: dict[Atom, Atom] = {}
    index_set = set(atom_indices)

    for chain in topology._chains:
        newChain = newTopology.add_chain()
        newChain.chain_id = chain.chain_id
        for residue in chain._residues:
            resSeq = getattr(residue, "resSeq", residue.index)
            newResidue = None
            for atom in residue._atoms:
                if atom.index in index_set:
                    try:  # OpenMM Topology objects don't have serial attributes, so we have to check first.
                        serial = atom.serial
                    except AttributeError:
                        serial = None
                    if newResidue is None:
                        newResidue = newTopology.add_residue(
                            residue.name,
                            newChain,
                            resSeq,
                            residue.segment_id,
                        )
                    newAtom = newTopology.add_atom(
                        atom.name,
                        atom.element,
                        newResidue,
                        serial=serial,
                        formal_charge=atom.formal_charge,
                    )
                    old_atom_to_new_atom[atom] = newAtom

    bondsiter: Iterable[Bond] | Callable[[], Iterable[Bond]] = topology.bonds
    if not hasattr(bondsiter, "__iter__"):
        bondsiter = bondsiter()

    for bond in bondsiter:
        try:
            atom1, atom2 = bond
            newTopology.add_bond(
                old_atom_to_new_atom[atom1],
                old_atom_to_new_atom[atom2],
                type=bond.type,
                order=bond.order,
            )
        except KeyError:
            pass
            # we only put bonds into the new topology if both of their partners
            # were indexed and thus HAVE a new atom

    # Delete empty residues
    newTopology._residues = [r for r in newTopology._residues if len(r._atoms) > 0]
    for chain in newTopology._chains:
        chain._residues = [r for r in chain._residues if len(r._atoms) > 0]

    # Delete empty chains
    newTopology._chains = [c for c in newTopology._chains if len(c._residues) > 0]
    # Re-set the numAtoms and numResidues
    newTopology._numAtoms = ilen(newTopology.atoms)
    newTopology._numResidues = ilen(newTopology.residues)

    # Reset the chain indices
    for i, chain in enumerate(newTopology.chains):
        chain.index = i
    # Reset the residue indices
    for i, res in enumerate(newTopology.residues):
        res.index = i

    return newTopology


class Topology:
    """Topology stores the topological information about a system.

    The structure of a Topology object is similar to that of a PDB file.
    It consists of a set of Chains (often but not always corresponding to
    polymer chains).  Each Chain contains a set of Residues, and each Residue
    contains a set of Atoms.  In addition, the Topology stores a list of which
    atom pairs are bonded to each other.

    Atom and residue names should follow the PDB 3.0 nomenclature for all
    molecules for which one exists.

    Attributes
    ----------
    chains : generator
        Iterator over all Chains in the Topology.
    residues : generator
        Iterator over all Residues in the Chain.
    atoms : generator
        Iterator over all Atoms in the Chain.
    bonds : generator
        Iterator over all Bonds in the Topology

    Examples
    --------
    >>> topology = md.load('example.pdb').topology
    >>> print(topology)
    <mdtraj.Topology with 1 chains, 3 residues, 22 atoms, 21 bonds at 0x105a98e90>
    >>> table, bonds = topology.to_dataframe()
    >>> print(table.head())
       serial name element  resSeq resName  chainID
    0       0   H1       H       1     CYS        0
    1       1  CH3       C       1     CYS        0
    2       2   H2       H       1     CYS        0
    3       3   H3       H       1     CYS        0
    4       4    C       C       1     CYS        0
    >>> # rename residue "CYS" to "CYSS"
    >>> table[table['residue'] == 'CYS']['residue'] = 'CYSS'
    >>> print(table.head())
       serial name element  resSeq resName   chainID
    0       0   H1       H       1     CYSS        0
    1       1  CH3       C       1     CYSS        0
    2       2   H2       H       1     CYSS        0
    3       3   H3       H       1     CYSS        0
    4       4    C       C       1     CYSS        0
    >>> t2 = md.Topology.from_dataframe(table, bonds)
    """

    _standardBonds: dict = {}

    def __init__(self) -> None:
        """Create a new Topology object"""
        self._chains: list[Chain] = []
        self._numResidues: int = 0
        self._numAtoms: int = 0
        self._bonds: list[Bond] = []
        self._atoms: list[Atom] = []
        self._residues: list[Residue] = []

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return "<%s>" % (self._string_summary_basic())

    def __repr__(self) -> str:
        return f"<{self._string_summary_basic()} at 0x{id(self):02x}>"

    def _string_summary_basic(self) -> str:
        return "mdtraj.Topology with %d chains, %d residues, %d atoms, %d bonds" % (
            self.n_chains,
            self.n_residues,
            self.n_atoms,
            len(self._bonds),
        )

    def copy(self) -> Topology:
        """Return a copy of the topology

        Returns
        -------
        out : Topology
            A copy of this topology
        """
        out = Topology()
        for chain in self.chains:
            c = out.add_chain()
            c.chain_id = chain.chain_id
            for residue in chain.residues:
                r = out.add_residue(residue.name, c, residue.resSeq, residue.segment_id)
                for atom in residue.atoms:
                    out.add_atom(atom.name, atom.element, r, serial=atom.serial, formal_charge=atom.formal_charge)

        for bond in self.bonds:
            a1, a2 = bond
            out.add_bond(a1, a2, type=bond.type, order=bond.order)

        return out

    def __copy__(self, *args) -> Topology:
        return self.copy()

    def __deepcopy__(self, *args) -> Topology:
        return self.copy()

    def __hash__(self) -> int:
        hash_value = hash(tuple(self._chains))
        hash_value ^= hash(tuple(self._atoms))
        hash_value ^= hash(tuple(self._bonds))
        hash_value ^= hash(tuple(self._residues))

        return hash_value

    def join(self, other: Topology, keep_resSeq: bool = True) -> Topology:
        """Join two topologies together

        Parameters
        ----------
        other : Topology
            Another topology object
        keep_resSeq : bool, optional, default=True
            if False the residue numbers (resSeq) of the
            topology that is joined are updated in order to
            continue from the last resSeq of this topology

        Returns
        -------
        out : Topology
            A joint topology, with all of the atoms/residues/chains/bonds
            in each of the individual topologies
        """
        if not isinstance(other, Topology):
            raise ValueError("other must be an instance of Topology to join")
        out = self.copy()

        # I need this in order to have the resSeq of the
        # new residues to continue from the one of the
        # las residue of this topology
        if not keep_resSeq:
            out_resSeq = out.atom(-1).residue.resSeq

        atom_mapping: dict[Atom, Atom] = {}
        for chain in other.chains:
            c = out.add_chain()
            for residue in chain.residues:
                if keep_resSeq:
                    out_resSeq = residue.resSeq
                else:
                    out_resSeq += 1
                r = out.add_residue(residue.name, c, out_resSeq, residue.segment_id)
                for atom in residue.atoms:
                    a = out.add_atom(atom.name, atom.element, r, serial=atom.serial, formal_charge=atom.formal_charge)
                    atom_mapping[atom] = a

        for bond in other.bonds:
            a1, a2 = bond
            out.add_bond(
                atom_mapping[a1],
                atom_mapping[a2],
                type=bond.type,
                order=bond.order,
            )

        return out

    def to_fasta(self, chain: int | None = None) -> str | list[str]:
        """Convert this topology into FASTA string

        Parameters
        ----------
        chain : Integer, optional, default=None
            If specified, will return the FASTA string for this chain in the
            Topology.

        Returns
        -------
        fasta : String or list of Strings
           A FASTA string for each chain specified.
        """

        def fasta(chain_obj):
            return "".join(
                res.code for res in chain_obj.residues if (res.is_protein or res.is_nucleic) and res.code is not None
            )

        if chain is not None:
            if not isinstance(chain, int):
                raise ValueError("chain must be an Integer.")
            return fasta(self._chains[chain])
        else:
            return [fasta(c) for c in self._chains]

    def to_openmm(self, traj: md.Trajectory | None = None) -> openmm.app.Topology:
        """Convert this topology into OpenMM topology

        Parameters
        ----------
        traj : MDTraj.Trajectory, optional, default=None
            If specified, use the first frame from this trajectory to
            set the unitcell information in the openmm topology.

        Returns
        -------
        topology : openmm.app.Topology
            This topology, as an OpenMM topology
        """
        app = import_("openmm.app")
        mm = import_("openmm")
        u = import_("openmm.unit")

        out = app.Topology()
        atom_mapping = {}
        bond_mapping = {
            Single: app.Single,
            Double: app.Double,
            Triple: app.Triple,
            Amide: app.Amide,
            Aromatic: app.Aromatic,
            None: None,
        }

        for chain in self.chains:
            c = out.addChain(id=chain.chain_id)
            for residue in chain.residues:
                r = out.addResidue(residue.name, c, id=str(residue.resSeq))
                for atom in residue.atoms:
                    if atom.element is elem.virtual:
                        element = None
                    else:
                        element = app.Element.getBySymbol(atom.element.symbol)

                    # If we are using a compatible version of OpenMM, add formal charge
                    try:
                        a = out.addAtom(atom.name, element, r, formalCharge=atom.formal_charge)
                    except (
                        TypeError
                    ):  # This will occur if the version of OpenMM does not support formal charge (version<=8.2)
                        a = out.addAtom(atom.name, element, r)
                    atom_mapping[atom] = a

        for bond in self.bonds:
            a1, a2 = bond
            out.addBond(
                atom_mapping[a1],
                atom_mapping[a2],
                type=bond_mapping[bond.type],
                order=bond.order,
            )

        if traj is not None:
            angles = traj.unitcell_angles[0]

            if np.linalg.norm(angles - 90.0) > 1e-4:
                raise (
                    ValueError(
                        "Unitcell angles must be 90.0 to use in OpenMM topology.",
                    )
                )

            box_vectors = mm.Vec3(*traj.unitcell_lengths[0]) * u.nanometer
            out.setUnitCellDimensions(box_vectors)

        return out

    @classmethod
    def from_openmm(cls, value: openmm.app.Topology) -> Topology:
        """Create a mdtraj topology from an OpenMM topology

        Parameters
        ----------
        value : openmm.app.Topology
            An OpenMM topology that you wish to convert to a
            mdtraj topology.
        """
        app = import_("openmm.app")

        bond_mapping = {
            app.Single: Single,
            app.Double: Double,
            app.Triple: Triple,
            app.Amide: Amide,
            app.Aromatic: Aromatic,
            None: None,
        }

        if not isinstance(value, app.Topology):
            raise TypeError(
                "value must be an OpenMM Topology. You supplied a %s" % type(value),
            )

        out = cls()
        atom_mapping = {}

        for chain in value.chains():
            c = out.add_chain(chain_id=chain.id)
            for residue in chain.residues():
                try:
                    r = out.add_residue(residue.name, c, resSeq=int(residue.id))
                except ValueError:
                    r = out.add_residue(residue.name, c)
                for atom in residue.atoms():
                    if atom.element is None:
                        element = elem.virtual
                    else:
                        element = elem.get_by_symbol(atom.element.symbol)
                    # If we are using a compatible version of OpenMM (>=8.2), add formal charge
                    if hasattr(atom, "formalCharge"):
                        a = out.add_atom(atom.name, element, r, formal_charge=atom.formalCharge)
                    else:
                        a = out.add_atom(atom.name, element, r)
                    atom_mapping[atom] = a

        for bond in value.bonds():
            a1, a2 = bond
            out.add_bond(
                atom_mapping[a1],
                atom_mapping[a2],
                type=bond_mapping[bond.type],
                order=bond.order,
            )

        return out

    def to_dataframe(self) -> tuple[pd.DataFrame, NDArray[np.float64] | None]:
        """Convert this topology into a pandas dataframe

        Returns
        -------
        atoms : pandas.DataFrame
            The atoms in the topology, represented as a data frame.
        bonds : np.ndarray, shape=(n_bonds, 4), dtype=float, Optional
            The bonds in this topology, represented as an n_bonds x 4 array indicating the two atom indices of the
            bond, the bond type, and bond order, cast to floats as the type is mapped from the classes to fractional.
            The atom indices and order are integers cast to float
        """
        pd = import_("pandas")
        data = [
            (
                atom.serial,
                atom.name,
                atom.element.symbol,
                atom.residue.resSeq,
                atom.residue.name,
                atom.residue.chain.index,
                atom.segment_id,
                atom.formal_charge,
            )
            for atom in self.atoms
        ]

        atoms = pd.DataFrame(
            data,
            columns=[
                "serial",
                "name",
                "element",
                "resSeq",
                "resName",
                "chainID",
                "segmentID",
                "formal_charge",
            ],
        )

        # Using Int64 makes the data type integer
        # but allows for NA values in the column
        atoms = atoms.astype({"formal_charge": "Int64"})

        bonds: NDArray[np.float64] = np.zeros([len(self._bonds), 4], dtype=float)
        for index, bond in enumerate(self.bonds):
            if bond.order is None:
                order = 0.0
            else:
                order = bond.order
            try:
                bond_type = float(bond.type) if bond.type is not None else 0.0
            except TypeError:
                # Trap the None case
                bond_type = 0.0
            bonds[index] = bond.atom1.index, bond.atom2.index, bond_type, order
        return atoms, bonds

    @classmethod
    def from_dataframe(cls, atoms: pd.DataFrame, bonds: NDArray[np.float64] | None = None) -> Topology:
        """Create a mdtraj topology from a pandas data frame

        Parameters
        ----------
        atoms : pandas.DataFrame
            The atoms in the topology, represented as a data frame. This data
            frame should have columns "serial" (atom index), "name" (atom name),
            "element" (atom's element), "resSeq" (index of the residue)
            "resName" (name of the residue), "chainID" (index of the chain),
            and optionally "segmentID" and "formal_charge", following the same conventions as wwPDB 3.0 format.
        bonds : np.ndarray, shape=(n_bonds, 4) or (n_bonds, 2), dtype=float, Optional
            The bonds in the topology, represented as a n_bonds x 4 or n_bonds x 2
            size array of the indices of the atoms involved, type, and order of each
            bond, represented as floats. Type and order are optional. Specifying bonds
            here is optional. To create standard protein bonds, you can use
            `create_standard_bonds` to "fill in" the bonds on your newly created
            Topology object, although type and order of bond are not computed if
            that method is used.

        See Also
        --------
        create_standard_bonds
        """
        pd = import_("pandas")

        if bonds is None:
            bonds = np.zeros([0, 4], dtype=float)

        for col in [
            "name",
            "element",
            "resSeq",
            "resName",
            "chainID",
            "serial",
        ]:
            if col not in atoms.columns:
                raise ValueError("dataframe must have column %s" % col)

        if "formal_charge" not in atoms.columns:
            atoms["formal_charge"] = None

        if "segmentID" not in atoms.columns:
            atoms["segmentID"] = ""

        # Ensure formal_charge is of type Int64 (nullable integer)
        try:
            atoms = atoms.astype({"formal_charge": "Int64"})
        except ValueError as e:
            raise ValueError(
                "Formal charge conversion failed: all values in 'formal_charge' must be integers "
                f"or None. Original error: {e}",
            )

        out = cls()
        if not isinstance(atoms, pd.DataFrame):
            raise TypeError(
                "atoms must be an instance of pandas.DataFrame. You supplied a %s" % type(atoms),
            )
        if not isinstance(bonds, np.ndarray):
            raise TypeError(
                "bonds must be an instance of numpy.ndarray. You supplied a %s" % type(bonds),
            )

        if not np.all(np.arange(len(atoms)) == atoms.index):
            raise ValueError(
                "atoms must be uniquely numbered starting from zero.",
            )

        out._atoms = [None for i in range(len(atoms))]  # type: ignore

        c = None
        r = None
        previous_chainID = None
        previous_resName = None
        previous_resSeq = None
        for atom_index, atom in atoms.iterrows():
            int(atom_index)  # Fixes bizarre hashing issue on Py3K.  See #545

            if atom["chainID"] != previous_chainID:
                previous_chainID = atom["chainID"]

                c = out.add_chain()

            if (
                atom["resSeq"] != previous_resSeq or atom["resName"] != previous_resName or c.n_atoms == 0  # type: ignore
            ):
                previous_resSeq = atom["resSeq"]
                previous_resName = atom["resName"]

                r = out.add_residue(
                    atom["resName"],
                    c,  # type: ignore
                    atom["resSeq"],
                    atom["segmentID"],
                )

            a = Atom(
                atom["name"],
                elem.get_by_symbol(atom["element"]),
                atom_index,
                r,  # type: ignore
                serial=atom["serial"],
                formal_charge=atom["formal_charge"] if pd.notna(atom["formal_charge"]) else None,  # Fix PD sentinel
            )
            out._atoms[atom_index] = a
            r._atoms.append(a)  # type: ignore

        for bond in bonds:
            ai1 = int(bond[0])
            ai2 = int(bond[1])
            try:
                bond_type = float_to_bond_type(bond[2])
                bond_order: int | None = int(bond[3])
                if bond_order == 0:
                    bond_order = None
            except IndexError:
                # Does not exist
                bond_type = None
                bond_order = None
            out.add_bond(out.atom(ai1), out.atom(ai2), bond_type, bond_order)

        out._numAtoms = out.n_atoms
        return out

    def to_bondgraph(self) -> nx.Graph:
        """Create a NetworkX graph from the atoms and bonds in this topology

        Returns
        -------
        g : nx.Graph
            A graph whose nodes are the Atoms in this topology, and
            whose edges are the bonds

        See Also
        --------
        atoms
        bonds

        Notes
        -----
        This method requires the NetworkX python package.
        """
        nx = import_("networkx")
        g = nx.Graph()
        g.add_nodes_from(self.atoms)
        g.add_edges_from(self.bonds)
        return g

    def __eq__(self, other: object) -> bool:
        """Are two topologies equal?

        Parameters
        ----------
        other : object
            The object to compare to

        Returns
        -------
        equality : bool
            Are the two topologies identical?
        """
        if not isinstance(other, Topology):
            return False
        if self is other:
            return True

        if len(self._chains) != len(other._chains):
            return False

        for c1, c2 in zip(self.chains, other.chains):
            if c1.index != c2.index:
                return False
            if len(c1._residues) != len(c2._residues):
                return False

            for r1, r2 in zip(c1.residues, c2.residues):
                if (r1.index != r1.index) or (r1.name != r2.name):  # or (r1.resSeq != r2.resSeq):
                    return False
                if len(r1._atoms) != len(r2._atoms):
                    return False

                for a1, a2 in zip(r1.atoms, r2.atoms):
                    if (a1.index != a2.index) or (a1.name != a2.name):
                        return False
                    if a1.element is not a2.element:
                        return False
                    # for attr in ['atomic_number', 'name', 'symbol']:
                    #     if getattr(a1.element, attr) != getattr(a2.element, attr):
                    #         return False

        if len(self._bonds) != len(other._bonds):
            return False

        # the bond ordering is somewhat ambiguous, so try and fix it for comparison
        self_sorted_bonds = sorted([bond for bond in self.bonds])
        other_sorted_bonds = sorted([bond for bond in other.bonds])

        for i, bond in enumerate(self_sorted_bonds):
            if bond != other_sorted_bonds[i]:
                return False

        return True

    def add_chain(self, chain_id: str | None = None) -> Chain:
        """Create a new Chain and add it to the Topology.

        Parameters
        ----------
        chain_id : str
            The PDB chainID of the Chain.

        Returns
        -------
        chain : mdtraj.topology.Chain
            the newly created Chain
        """
        chain = Chain(len(self._chains), self, chain_id)
        self._chains.append(chain)
        return chain

    def add_residue(self, name: str, chain: Chain, resSeq: int | None = None, segment_id: str = "") -> Residue:
        """Create a new Residue and add it to the Topology.

        Parameters
        ----------
        name : str
            The name of the residue to add
        chain : mdtraj.topology.Chain
            The Chain to add it to
        resSeq : int, optional
            Residue sequence number, such as from a PDB record. These sequence
            numbers are arbitrary, and do not necessarily start at 0 (or 1).
            If not supplied, the resSeq attribute will be set to the
            residue's sequential (0 based) index.
        segment_id : str, optional
            A label for the segment to which this residue belongs

        Returns
        -------
        residue : mdtraj.topology.Residue
            The newly created Residue
        """
        if resSeq is None:
            resSeq = self._numResidues
        residue = Residue(name, self._numResidues, chain, resSeq, segment_id)
        self._residues.append(residue)
        self._numResidues += 1
        chain._residues.append(residue)
        return residue

    def insert_atom(
        self,
        name: str,
        element: md.element.Element,
        residue: Residue,
        index: int | None = None,
        rindex: int | None = None,
        serial: int | None = None,
    ) -> Atom:
        """Create a new Atom and insert it into the Topology at a specific position.

        Parameters
        ----------
        name : str
            The name of the atom to add
        element : mdtraj.element.Element
            The element of the atom to add
        residue : mdtraj.topology.Residue
            The Residue to add it to
        index : int
            If provided, the desired index for this atom
            within the topology. Existing atoms with
            indices >= index will be pushed back.
        rindex : int
            If provided, the desired position for this atom
            within the residue
        serial : int
            Serial number associated with the atom.
            This has nothing to do with the actual ordering
            and is solely for PDB labeling purposes.

        Returns
        -------
        atom : mdtraj.topology.Atom
            the newly created Atom
        """
        if element is None:
            element = elem.virtual
        if index is None:
            atom = Atom(name, element, self._numAtoms, residue, serial=serial)
            self._atoms.append(atom)
        else:
            atom = Atom(name, element, index, residue, serial=serial)
            for i in range(index, len(self._atoms)):
                self._atoms[i].index += 1
            self._atoms.insert(index, atom)
        self._numAtoms += 1
        if rindex is None:
            residue._atoms.append(atom)
        else:
            residue._atoms.insert(rindex, atom)
        return atom

    def delete_atom_by_index(self, index: int) -> None:
        """Delete an Atom from the topology.
        Atom is removed and all subsequent atoms are reindexed.
        All bonds involving this atom are likewise removed. This may result
        in disconnected groups of atoms.

        Parameters
        ----------
        index : int
            The index of the atom to be removed.
        """
        a = self._atoms[index]
        assert a is not None, "Expected an Atom, but got None"
        if a.index != index:
            raise RuntimeError(
                "Index of selected atom does not match order in topology.",
            )
        for i in range(index + 1, len(self._atoms)):
            self._atoms[i].index -= 1

        self._bonds = [bond for bond in self._bonds if a not in bond]

        a.residue._atoms.remove(a)
        self._atoms.remove(a)
        self._numAtoms -= 1

    def add_atom(
        self,
        name: str,
        element: md.element.Element,
        residue: Residue,
        serial: int | None = None,
        formal_charge: int | None = None,
    ) -> Atom:
        """Create a new Atom and add it to the Topology.

        Parameters
        ----------
        name : str
            The name of the atom to add
        element : mdtraj.element.Element
            The element of the atom to add
        residue : mdtraj.topology.Residue
            The Residue to add it to
        serial : int, optional
            Serial number associated with the atom.
        formal_charge : int, optional
            The formal charge of the atom to add.

        Returns
        -------
        atom : mdtraj.topology.Atom
            the newly created Atom
        """
        if element is None:
            element = elem.virtual
        atom = Atom(
            name,
            element,
            self._numAtoms,
            residue,
            serial=serial,
            formal_charge=formal_charge,
        )
        self._atoms.append(atom)
        self._numAtoms += 1
        residue._atoms.append(atom)
        return atom

    def add_bond(
        self,
        atom1: Atom,
        atom2: Atom,
        type: Singleton | None = None,
        order: int | None = None,
    ) -> None:
        """Create a new bond and add it to the Topology.

        Parameters
        ----------
        atom1 : mdtraj.topology.Atom
            The first Atom connected by the bond
        atom2 : mdtraj.topology.Atom
            The second Atom connected by the bond
        type : mdtraj.topology.Singleton or None, Default: None, Optional
            Bond type of the bond, or None if not known/provided
        order : 1, 2, 3 or None, Default: None, Optional
            Characteristic order of the bond if known, defaults None
        """
        if atom1.index < atom2.index:
            self._bonds.append(Bond(atom1, atom2, type=type, order=order))
        else:
            self._bonds.append(Bond(atom2, atom1, type=type, order=order))

    def chain(self, index: int) -> Chain:
        """Get a specific chain by index.  These indices
        start from zero.

        Parameters
        ----------
        index : int
            The index of the chain to select.

        Returns
        -------
        chain : Chain
            The `index`-th chain in the topology.
        """
        return self._chains[index]

    @property
    def chains(self) -> Iterator[Chain]:
        """Iterator over all Chains in the Topology.

        Returns
        -------
        chainiter : listiterator
            Iterator over all Chains in the Topology.
        """
        return iter(self._chains)

    @property
    def n_chains(self) -> int:
        """Get the number of chains in the Topology"""
        return len(self._chains)

    def residue(self, index: int) -> Residue:
        """Get a specific residue by index.  These indices
        start from zero.

        Parameters
        ----------
        index : int
            The index of the residue to select.

        Returns
        -------
        residue : Residue
            The `index`-th residue in the topology.
        """
        return self._residues[index]

    @property
    def residues(self) -> Iterator[Residue]:
        """Iterator over all Residues in the Topology.

        Returns
        -------
        residueiter : generator
            Iterator over all Residues in the Topology.
        """
        for chain in self._chains:
            yield from chain._residues

    @property
    def n_residues(self) -> int:
        """Get the number of residues in the Topology."""
        return len(self._residues)

    def atom(self, index: int) -> Atom:
        """Get a specific atom by index. These indices
        start from zero.

        Parameters
        ----------
        index : int
            The index of the atom to select.

        Returns
        -------
        atom : Atom
            The `index`-th atom in the topology.
        """

        return self._atoms[index]

    @property
    def atoms(self) -> Iterator[Atom]:
        """Iterator over all Atoms in the Topology.

        Returns
        -------
        atomiter : generator
            Iterator over all Atoms in the Topology.
        """
        for chain in self._chains:
            for residue in chain._residues:
                yield from residue._atoms

    def atoms_by_name(self, name: str) -> Iterator[Atom]:
        """Iterator over all Atoms in the Topology with a specified name

        Parameters
        ----------
        name : str
            The particular atom name of interest.

        Examples
        --------
        >>> for atom in topology.atoms_by_name('CA'):
        ...     print(atom)

        Returns
        -------
        atomiter : generator
        """
        for atom in self.atoms:
            if atom.name == name:
                yield atom

    @property
    def n_atoms(self) -> int:
        """Get the number of atoms in the Topology"""
        return len(self._atoms)

    @property
    def bonds(self) -> Iterator[Bond]:
        """Iterator over all bonds (each represented as a tuple of two Atoms) in the Topology.

        Returns
        -------
        bonditer : generator
            Iterator over all bonds between Atoms in the Trajectory.
            Each bond can then be iterated to get the atoms.
            I.e.: `for atom1, atom2 in Topology.bonds` yields iterator over pairs of Atoms
                  `for bond in Topology.bonds` yields iterator over bonds
        """
        return iter(self._bonds)

    @property
    def n_bonds(self) -> int:
        """Get the number of bonds in the Topology"""
        return len(self._bonds)

    def create_standard_bonds(self) -> None:
        """Create bonds based on the atom and residue names for all standard residue types."""
        if len(Topology._standardBonds) == 0:
            # Load the standard bond defitions.

            tree = etree.parse(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "formats",
                    "pdb",
                    "data",
                    "residues.xml",
                ),
            )
            for residue in tree.getroot().findall("Residue"):
                bonds: list[tuple[str, str]] = []
                Topology._standardBonds[residue.attrib["name"]] = bonds
                for bond in residue.findall("Bond"):
                    bonds.append((bond.attrib["from"], bond.attrib["to"]))
        for chain in self._chains:
            # First build a map of atom names to atoms.

            atomMaps = []
            for residue in chain._residues:
                atomMap: dict[str, Atom] = {}
                atomMaps.append(atomMap)
                for atom in residue._atoms:
                    atomMap[atom.name] = atom

            # Loop over residues and construct bonds.

            for i in range(len(chain._residues)):
                name = chain._residues[i].name
                if name in Topology._standardBonds:
                    for bond in Topology._standardBonds[name]:
                        if bond[0].startswith("-") and i > 0:
                            fromResidue = i - 1
                            fromAtom = bond[0][1:]
                        elif bond[0].startswith("+") and i < len(chain._residues):
                            fromResidue = i + 1
                            fromAtom = bond[0][1:]
                        else:
                            fromResidue = i
                            fromAtom = bond[0]
                        if bond[1].startswith("-") and i > 0:
                            toResidue = i - 1
                            toAtom = bond[1][1:]
                        elif bond[1].startswith("+") and i < len(chain._residues):
                            toResidue = i + 1
                            toAtom = bond[1][1:]
                        else:
                            toResidue = i
                            toAtom = bond[1]
                        if fromAtom in atomMaps[fromResidue] and toAtom in atomMaps[toResidue]:
                            self.add_bond(
                                atomMaps[fromResidue][fromAtom],
                                atomMaps[toResidue][toAtom],
                            )

    def create_disulfide_bonds(self, positions: list) -> None:
        """Identify disulfide bonds based on proximity and add them to the Topology.

        Parameters
        ----------
        positions : list
            The list of atomic positions based on which to identify bonded atoms
        """

        def isCyx(res):
            names = [atom.name for atom in res._atoms]
            return "SG" in names and "HG" not in names

        cyx = [res for res in self.residues if res.name == "CYS" and isCyx(res)]
        atomNames = [[atom.name for atom in res._atoms] for res in cyx]
        for i in range(len(cyx)):
            sg1 = cyx[i]._atoms[atomNames[i].index("SG")]
            pos1 = positions[sg1.index]
            for j in range(i):
                sg2 = cyx[j]._atoms[atomNames[j].index("SG")]
                pos2 = positions[sg2.index]
                delta = [x - y for (x, y) in zip(pos1, pos2)]
                distance = np.sqrt(
                    delta[0] * delta[0] + delta[1] * delta[1] + delta[2] * delta[2],
                )
                if distance < 0.3:  # this is supposed to be nm. I think we're good
                    self.add_bond(sg1, sg2)

    def subset(self, atom_indices: list[int]) -> Topology:
        """Create a new Topology from a subset of the atoms in an existing topology.

        Notes
        -----
        The existing topology will not be altered.

        Parameters
        ----------
        atom_indices : array_like
            A list of the indices corresponding to the atoms in that you'd
            like to retain.
        """
        return _topology_from_subset(self, atom_indices)

    def select_expression(self, selection_string: str) -> str:
        """Translate a atom selection expression into a pure python expression.

        Parameters
        ----------
        selection_string : str
            An expression in the MDTraj atom selection DSL

        Examples
        --------
        >>> topology.select_expression('name O and water')
        "[atom.index for atom in topology.atoms if ((atom.name == 'O') and atom.residue.is_water)]")

        Returns
        -------
        python_string : str
            A string containing a pure python expression, equivalent to the
            selection expression.
        """
        condition = parse_selection(selection_string).source
        fmt_string = "[atom.index for atom in topology.atoms if {condition}]"
        return fmt_string.format(condition=condition)

    def select(self, selection_string: str) -> NDArray[int]:
        """Execute a selection against the topology

        Parameters
        ----------
        selection_string : str
            An expression in the MDTraj atom selection DSL

        Examples
        --------
        >>> topology.select('name O and water')
        array([1, 3, 5, 10, ...])

        Returns
        -------
        indices : np.ndarray, dtype=int, ndim=1
            Array of the indices of the atoms matching the selection expression.

        See Also
        --------
        select_expression, mdtraj.core.selection.parse_selection
        """

        filter_func = parse_selection(selection_string).expr  # type: ignore
        indices = np.array([a.index for a in self.atoms if filter_func(a)])
        return indices

    def select_atom_indices(
        self,
        selection: str = "minimal",
    ) -> NDArray[int]:
        """Get the indices of biologically-relevant groups by name.

        Parameters
        ----------
        selection : {'all', 'alpha', 'minimal', 'heavy', 'water'}
            What types of atoms to select.

            ``all``
                All atoms
            ``alpha``
                Protein residue alpha carbons
            ``minimal``
                Keep the atoms in protein residues with names in {CA, CB, C, N, O}
            ``heavy``
                All non-hydrogen protein atoms.
            ``water``
                Water oxygen atoms

        Returns
        ----------
        indices : np.ndarray (N,)
            An array of the indices of the selected atoms.
        """
        selection = selection.lower()
        options = ["all", "alpha", "minimal", "heavy", "water"]
        if selection == "all":
            atom_indices = np.arange(self.n_atoms)
        elif selection == "alpha":
            atom_indices = np.array(
                [a.index for a in self.atoms if a.name == "CA" and a.residue.is_protein],
                dtype=int,
            )
        elif selection == "minimal":
            atom_indices = np.array(
                [a.index for a in self.atoms if a.name in ["CA", "CB", "C", "N", "O"] and a.residue.is_protein],
                dtype=int,
            )
        elif selection == "heavy":
            atom_indices = np.array(
                [a.index for a in self.atoms if a.element != elem.hydrogen and a.residue.is_protein],
                dtype=int,
            )
        elif selection == "water":
            atom_indices = np.array(
                [a.index for a in self.atoms if a.name in ["O", "OW"] and a.residue.is_water],
                dtype=int,
            )
        else:
            raise ValueError(
                "{} is not a valid option. Selection must be one of {}".format(
                    selection,
                    ", ".join(options),
                ),
            )

        # indices = np.array(atom_indices)
        return atom_indices

    def select_pairs(
        self,
        selection1: str | NDArray[int] | Sequence[int] | None = None,
        selection2: str | NDArray[int] | Sequence[int] | None = None,
    ) -> NDArray[int]:
        """Generate unique pairs of atom indices.

        If a selection is a string, it will be resolved using the atom selection
        DSL, otherwise it is expected to be an array of atom indices.

        Parameters
        ----------
        selection1 : str or array-like, shape=(n_indices, ), dtype=int
            A selection for `select()` or an array of atom indices.
        selection2 : str or array-like, shape=(n_indices, ), dtype=int
            A selection for `select()` or an array of atom indices.

        Returns
        -------
        pairs : array-like, shape=(n_pairs, 2), dtype=int
            Each row gives the indices of two atoms.

        """
        # Resolve selections using the atom selection DSL...
        if isinstance(selection1, str):
            a_indices = self.select(selection1)
        else:  # ...or use a provided array of indices.
            a_indices = ensure_type(
                selection1,
                dtype=int,
                ndim=1,
                name="a_indices",
                warn_on_cast=False,
            )
        if isinstance(selection2, str):
            b_indices = self.select(selection2)
        else:
            b_indices = ensure_type(
                selection2,
                dtype=int,
                ndim=1,
                name="b_indices",
                warn_on_cast=False,
            )
        a_indices.sort()
        b_indices.sort()

        # Create unique pairs from the indices.
        # In the cases where a_indices and b_indices are identical or mutually
        # exclusive, we can utilize a more efficient and memory friendly
        # approach by removing the intermediate set creation required in
        # the general case.
        if np.array_equal(a_indices, b_indices):
            pairs = self._unique_pairs_equal(a_indices)
        elif len(np.intersect1d(a_indices, b_indices)) == 0:
            pairs = self._unique_pairs_mutually_exclusive(a_indices, b_indices)
        else:
            pairs = self._unique_pairs(a_indices, b_indices)
        return pairs

    @classmethod
    def _unique_pairs(cls, a_indices: Iterator[int], b_indices: Iterator[int]) -> NDArray[int]:
        return np.array(
            list(
                {(a, b) if a > b else (b, a) for a, b in itertools.product(a_indices, b_indices) if a != b},
            ),
            dtype=int,
        )

    @classmethod
    def _unique_pairs_mutually_exclusive(cls, a_indices: list[int], b_indices: list[int]) -> NDArray[int]:
        pairs = np.fromiter(
            itertools.chain.from_iterable(
                itertools.product(a_indices, b_indices),
            ),
            dtype=int,
            count=len(a_indices) * len(b_indices) * 2,
        )
        return np.vstack((pairs[::2], pairs[1::2])).T

    @classmethod
    def _unique_pairs_equal(cls, a_indices: list[int]) -> NDArray[int]:
        pairs = np.fromiter(
            itertools.chain.from_iterable(
                itertools.combinations(a_indices, 2),
            ),
            dtype=int,
            count=len(a_indices) * (len(a_indices) - 1),
        )
        return np.vstack((pairs[::2], pairs[1::2])).T

    def find_molecules(self) -> list[set[Atom]]:
        """Identify molecules based on bonds.

        A molecule is defined as a set of atoms that are connected to each other by bonds.
        This method uses the list of bonds to divide up the Topology's atoms into molecules.

        Returns
        -------
        molecules : list of sets
            Each entry represents one molecule, and is the set of all Atoms in that molecule
        """
        if len(self._bonds) == 0 and any(res.n_atoms > 1 for res in self._residues):
            raise ValueError(
                "Cannot identify molecules because this Topology does not include bonds",
            )

        # Make a list of every other atom to which each atom is connected.

        num_atoms = self.n_atoms
        atom_bonds: list[list[int]] = [[] for i in range(num_atoms)]
        for atom1, atom2 in self.bonds:
            atom_bonds[atom1.index].append(atom2.index)
            atom_bonds[atom2.index].append(atom1.index)

        # This is essentially a recursive algorithm, but it is reformulated as a loop to avoid
        # stack overflows.  It selects an atom, marks it as a new molecule, then recursively
        # marks every atom bonded to it as also being in that molecule.

        atom_molecule = [-1] * num_atoms
        num_molecules = 0
        for i in range(num_atoms):
            if atom_molecule[i] == -1:
                # Start a new molecule.

                atom_stack = [i]
                neighbor_stack = [0]
                molecule = num_molecules
                num_molecules += 1

                # Recursively tag all the bonded atoms.

                while len(atom_stack) > 0:
                    atom_index = atom_stack[-1]
                    atom_molecule[atom_index] = molecule
                    while (
                        neighbor_stack[-1] < len(atom_bonds[atom_index])
                        and atom_molecule[atom_bonds[atom_index][neighbor_stack[-1]]] != -1
                    ):
                        neighbor_stack[-1] += 1
                    if neighbor_stack[-1] < len(atom_bonds[atom_index]):
                        atom_stack.append(atom_bonds[atom_index][neighbor_stack[-1]])
                        neighbor_stack.append(0)
                    else:
                        del atom_stack[-1]
                        del neighbor_stack[-1]

        # Build the final output.

        molecules: list[set[Atom]] = [set() for i in range(num_molecules)]
        for atom in self.atoms:
            molecules[atom_molecule[atom.index]].add(atom)
        return molecules

    def guess_anchor_molecules(self) -> list[set[Atom]]:
        """Guess anchor molecules for imaging

        Returns
        -------
        anchor_molecules : list of atom sets
            List of anchor molecules

        See Also
        --------
        Trajectory.image_molecules
        """
        molecules = self.find_molecules()

        # Select the anchor molecules.
        molecules.sort(key=lambda x: -len(x))
        atoms_cutoff = max(
            len(molecules[int(0.1 * len(molecules))]),
            int(0.1 * len(molecules[0])),
        )
        anchor_molecules = [mol for mol in molecules if len(mol) > atoms_cutoff]
        num_anchors = len(anchor_molecules)
        if num_anchors == 0:
            raise ValueError(
                "Could not find any anchor molecules. Based on "
                "our heuristic, those should be molecules with "
                f"more than {atoms_cutoff} atoms. Perhaps your topology "
                "doesn't give an accurate bond graph?",
            )
        return anchor_molecules


class Chain:
    """A Chain object represents a chain within a Topology.

    Attributes
    ----------
    index : int
        The index of the Chain within its Topology
    topology : mdtraj.Topology
        The Topology this Chain belongs to
    chain_id : str
        The PDB chainID of the Chain.
    residues : generator
        Iterator over all Residues in the Chain.
    atoms : generator
        Iterator over all Atoms in the Chain.
    """

    def __init__(self, index: int, topology: Topology, chain_id: str | None = None) -> None:
        """Construct a new Chain.  You should call add_chain() on the Topology instead of calling this directly."""
        # The index of the Chain within its Topology
        self.index: int = index
        # The Topology this Chain belongs to
        self.topology: Topology = topology
        self._residues: list[Residue] = []
        # PDB format chainID
        self.chain_id: str | None = chain_id

    @property
    def residues(self) -> Iterator[Residue]:
        """Iterator over all Residues in the Chain.

        Returns
        -------
        residueiter : listiterator
            Iterator over all Residues in the Topology.
        """
        return iter(self._residues)

    def residue(self, index: int) -> Residue:
        """Get a specific residue in this Chain.

        Parameters
        ----------
        index : int
            The index of the residue to select.

        Returns
        -------
        residue : Residue
        """
        return self._residues[index]

    @property
    def n_residues(self) -> int:
        """Get the number of residues in this Chain."""
        return len(self._residues)

    @property
    def atoms(self) -> Iterator[Atom]:
        """Iterator over all Atoms in the Chain.

        Returns
        -------
        atomiter : generator
            Iterator over all Atoms in the Chain.
        """
        for residue in self._residues:
            yield from residue._atoms

    def atoms_by_name(self, name: str) -> Iterator[Atom]:
        """Iterator over all Atoms in the Chain with a specified name.

        Parameters
        ----------
        name : str
            The particular atom name of interest.

        Examples
        --------
        >>> for atom in chain.atoms_by_name('CA'):
        ...     print(atom)

        Returns
        -------
        atomiter : generator
        """
        for atom in self.atoms:
            if atom.name == name:
                yield atom

    def atom(self, index: int) -> Atom:
        """Get a specific atom in this Chain.

        Parameters
        ----------
        index : int
            The index of the atom to select.

        Returns
        -------
        atom : Atom
        """
        # this could be made faster by caching the list
        # of atoms internally if necessary
        return next(itertools.islice(self.atoms, index, index + 1))

    @property
    def n_atoms(self) -> int:
        """Get the number of atoms in this Chain"""
        return sum(r.n_atoms for r in self._residues)

    def __hash__(self) -> int:
        return self.index


class Residue:
    """A Residue object represents a residue within a Topology.

    Attributes
    ----------
    name : str
        The name of the Residue
    index : int
        The index of the Residue within its Topology
    chain : mdtraj.topology.Chain
        The chain within which this residue belongs
    resSeq : int
        The residue sequence number
    segment_id : str, optional
        A label for the segment to which this residue belongs
    """

    def __init__(self, name: str, index: int, chain: Chain, resSeq: int, segment_id: str = "") -> None:
        """Construct a new Residue.  You should call add_residue()
        on the Topology instead of calling this directly."""
        self.name: str = name
        self.index: int = index
        self.chain: Chain = chain
        self.resSeq: int = resSeq
        self.segment_id: str = segment_id
        self._atoms: list[Atom] = []

    @property
    def atoms(self) -> Iterator[Atom]:
        """Iterator over all Atoms in the Residue.

        Returns
        -------
        atomiter : listiterator
            Iterator over all Atoms in the Residue.
        """
        return iter(self._atoms)

    def atoms_by_name(self, name: str) -> Iterator[Atom]:
        """Iterator over all Atoms in the Residue with a specified name

        Parameters
        ----------
        name : str
            The particular atom name of interest.

        Examples
        --------
        >>> for atom in residue.atoms_by_name('CA'):
        ...     print(atom)

        Returns
        -------
        atomiter : generator
        """
        for atom in self.atoms:
            if atom.name == name:
                yield atom

    def atom(self, index_or_name: int | str) -> Atom:
        """Get a specific atom in this Residue.

        Parameters
        ----------
        index_or_name : {int, str}
            Either a (zero-based) index, or the name of the atom. If a string
            is passed in, the first atom -- in index order -- with a matching
            name wil be returned.

        Returns
        -------
        atom : Atom
        """
        if isinstance(index_or_name, int):
            return self._atoms[index_or_name]
        else:
            try:
                return next(self.atoms_by_name(index_or_name))
            except StopIteration:
                raise KeyError("no matching atom found")

    @property
    def n_atoms(self) -> int:
        """Get the number of atoms in this Residue"""
        return len(self._atoms)

    @property
    def is_protein(self) -> bool:
        """Whether the residue is one found in proteins."""
        return self.name in _PROTEIN_RESIDUES

    @property
    def code(self) -> str | None:
        """Get the one letter code for this Residue"""
        if self.is_protein:
            return _AMINO_ACID_CODES[self.name]
        elif self.is_nucleic:
            return _NUCLEIC_ACID_CODES[self.name]
        else:
            return None

    @property
    def is_water(self) -> bool:
        """Whether the residue is water.

        Residue names according to VMD

        References
        ----------
        http://www.ks.uiuc.edu/Research/vmd/vmd-1.3/ug/node133.html
        """
        return self.name in _WATER_RESIDUES

    @property
    def is_nucleic(self) -> bool:
        """Whether the residue is one found in nucleic acids."""
        return self.name in _NUCLEIC_RESIDUES

    def __str__(self) -> str:
        return f"{self.name}{self.resSeq}"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash((self.name, self.index, self.resSeq, self.segment_id))


class Atom:
    """An Atom object represents a residue within a Topology.

    Attributes
    ----------
    name : str
        The name of the Atom
    element : mdtraj.element.Element
        The element of the Atoms
    index : int
        The index of the Atom within its Topology
    residue : mdtraj.topology.Residue
        The Residue this Atom belongs to
    serial : int
        The serial number from the PDB specification. Unlike index,
        this may not be contiguous or 0-indexed.
    formal_charge : float

    """

    def __init__(
        self,
        name: str,
        element: md.element.Element,
        index: int,
        residue: Residue,
        serial: int | None = None,
        formal_charge: int | None = None,
    ):
        """Construct a new Atom.  You should call add_atom() on the Topology instead of calling this directly."""
        # The name of the Atom
        self.name: str = name
        # That Atom's element
        self.element: md.element.Element = element
        # The index of the Atom within its Topology
        self.index: int = index
        # The Residue this Atom belongs to
        self.residue: Residue = residue
        # The not-necessarily-contiguous "serial" number from the PDB spec
        self.serial: int | None = serial
        # The formal charge of the atom
        self.formal_charge: int | None = formal_charge

    @property
    def n_bonds(self) -> int:
        """Number of bonds in which the atom participates."""
        # TODO: this info could be cached.
        return ilen(bond for bond in self.residue.chain.topology.bonds if self in bond)

    @property
    def is_backbone(self) -> bool:
        """Whether the atom is in the backbone of a protein residue"""
        return self.name in {"C", "CA", "N", "O"} and self.residue.is_protein

    @property
    def is_sidechain(self) -> bool:
        """Whether the atom is in the sidechain of a protein residue"""
        return self.name not in {"C", "CA", "N", "O", "HA", "H"} and self.residue.is_protein

    @property
    def segment_id(self) -> str:
        """User specified segment_id of the residue to which this atom belongs"""
        return self.residue.segment_id

    def __eq__(self, other: object) -> bool:
        """Check whether two Atom objects are equal."""
        if not isinstance(other, Atom):
            return NotImplemented
        if self.name != other.name:
            return False
        if self.index != other.index:
            return False
        if self.formal_charge != other.formal_charge:
            return False
        if self.element.name != other.element.name:
            return False
        if self.residue.name != other.residue.name:
            return False
        if self.residue.index != other.residue.index:
            return False
        if self.residue.chain.index != other.residue.chain.index:
            return False
        return True

    def __hash__(self) -> int:
        """A quick comparison."""
        return self.index

    def __str__(self) -> str:
        return f"{self.residue}-{self.name}"

    def __repr__(self) -> str:
        return str(self)


# Enumerated values for bond type


class Single(Singleton):
    def __repr__(self) -> str:
        return "Single"

    def __float__(self) -> float:
        return 1.0


Single: Singleton = Single()  # type: ignore[no-redef]


class Double(Singleton):
    def __repr__(self) -> str:
        return "Double"

    def __float__(self) -> float:
        return 2.0


Double: Singleton = Double()  # type: ignore[no-redef]


class Triple(Singleton):
    def __repr__(self) -> str:
        return "Triple"

    def __float__(self) -> float:
        return 3.0


Triple: Singleton = Triple()  # type: ignore[no-redef]


class Aromatic(Singleton):
    def __repr__(self) -> str:
        return "Aromatic"

    def __float__(self) -> float:
        return 1.5


Aromatic: Singleton = Aromatic()  # type: ignore[no-redef]


class Amide(Singleton):
    def __repr__(self) -> str:
        return "Amide"

    def __float__(self) -> float:
        return 1.25


Amide: Singleton = Amide()  # type: ignore[no-redef]


def float_to_bond_type(bond_float: float | None) -> Singleton | None:
    """
    Convert a float to known bond type class, or None if no matched class if found

    Parameters
    ----------
    bond_float : float or None
        Representation of built in bond types as a float,
        Maps this float to specific bond class, if the float has no map, None is returned instead

    Returns
    -------
    bond_type : mdtraj.topology.Singleton subclass or None
        Bond type matched to the float if known
        If no match is found, returns None (which is also a valid type for the Bond class)
    """
    # Explicitly cast to Singleton to avoid mypy error
    bond_mapping: dict[float, Singleton] = {
        1.0: cast(Singleton, Single),
        2.0: cast(Singleton, Double),
        3.0: cast(Singleton, Triple),
        1.5: cast(Singleton, Aromatic),
        1.25: cast(Singleton, Amide),
    }

    try:
        return bond_mapping.get(float(bond_float), None)
    except (TypeError, ValueError):
        # Skip cases where you cannot cast to float (e.g. float(None))
        return None


def bond_name_to_type(name: str) -> Singleton | None:
    """
    Convert a bond name to known bond type class, or None if no matched class if found

    Parameters
    ----------
    name : str
        Name of the bond type, e.g. 'single', 'double', 'aromatic', etc.

    Returns
    -------
    bond_type : mdtraj.topology.Singleton subclass or None
        Bond type matched to the name if known
        If no match is found, returns None (which is also a valid type for the Bond class)
    """

    # Explicitly cast to Singleton to avoid mypy error
    bond_mapping: dict[str, Singleton] = {
        "single": cast(Singleton, Single),
        "double": cast(Singleton, Double),
        "triple": cast(Singleton, Triple),
        "aromatic": cast(Singleton, Aromatic),
        "amide": cast(Singleton, Amide),
    }

    try:
        # normalize name
        name = name.lower().strip()
        return bond_mapping.get(name)
    except (TypeError, AttributeError):
        # skip cases where you cannot normalize (e.g. None)
        return None


class Bond(namedtuple("Bond", ["atom1", "atom2"])):
    # Add type annotations for the extra attributes.
    type: Singleton | None
    order: int | None
    """A Bond object represents a bond between two Atoms within a Topology.

    Attributes
    ----------
    atom1 : mdtraj.topology.Atom
        The first atom in the bond
    atom2 : mdtraj.topology.Atom
        The second atom in the bond
    order : instance of mdtraj.topology.Singleton or None
    type : int on [1,3] domain or None
    """

    def __new__(
        cls,
        atom1: Atom,
        atom2: Atom,
        type: Singleton | None,
        order: int | None,
    ) -> Bond:
        """Construct a new Bond.  You should call add_bond()
        on the Topology instead of calling this directly.

        Must use __new__ constructor since this is an immutable class
        """
        bond = super().__new__(cls, atom1, atom2)
        assert isinstance(type, Singleton) or type is None, "Type must be None or a Singleton"
        assert order is None or 1 <= order <= 3, "Order must be int between 1 to 3 or None"
        bond.type = type
        bond.order = order
        return bond

    def __getnewargs__(
        self,
    ) -> tuple[Atom, Atom, Singleton | None, int | None]:
        """
        Support for pickle protocol 2:
        http://docs.python.org/2/library/pickle.html#pickling-and-unpickling-normal-class-instances
        """
        return self[0], self[1], self.type, self.order

    @property
    def _equality_tuple(self) -> tuple[int, int, float, int]:
        # Hierarchy of parameters: Atom1 index -> Atom2 index -> type -> order
        return (
            self[0].index,
            self[1].index,
            float(self.type) if self.type is not None else 0.0,
            self.order if self.order is not None else 0,
        )

    def __deepcopy__(self, memo) -> Bond:
        return Bond(self[0], self[1], self.type, self.order)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bond):
            return NotImplemented
        return self._equality_tuple == other._equality_tuple

    def __repr__(self) -> str:
        s = f"Bond({self[0]}, {self[1]}"
        if self.type is not None:
            s = f"{s}, type={self.type}"
        if self.order is not None:
            s = "%s, order=%d" % (s, self.order)
        s += ")"
        return s

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        # Set of atoms making up bonds, the type, and the order
        return hash((self[0], self[1], self.type, self.order))

    @staticmethod
    def _other_is_bond(other):
        # Ensure other type for inequalities is a bond
        if not isinstance(other, Bond):
            raise TypeError("Bond inequalities can only be compared with other bonds")

    def __gt__(self, other) -> bool:
        # Cannot use total_ordering because namedtuple
        # has its own __gt__, __lt__, etc. methods, which
        # supersede total_ordering
        self._other_is_bond(other)
        return self._equality_tuple > other._equality_tuple

    def __ge__(self, other) -> bool:
        self._other_is_bond(other)
        return self._equality_tuple >= other._equality_tuple

    def __lt__(self, other) -> bool:
        self._other_is_bond(other)
        return self._equality_tuple < other._equality_tuple

    def __le__(self, other) -> bool:
        self._other_is_bond(other)
        return self._equality_tuple <= other._equality_tuple

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __getstate__(self) -> dict:
        # This is required for pickle because the parent class
        # does not properly return a state
        return self.__dict__



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Jeremy MG Leung
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import numpy as np
import pytest

from mdtraj.utils.unitcell import check_valid_unitcell_angles


class Test_Unitcell:
    @pytest.mark.parametrize(["alpha", "beta", "gamma"], [[90, 90, 90], [70, 90, 30]], ids=["orthogonal", "triclinic"])
    def test_valid(self, alpha, beta, gamma):
        """Test for valid unitcell angles, single frame"""
        unitcell_angles = np.array([alpha, beta, gamma], dtype=np.float32)

        assert check_valid_unitcell_angles(*unitcell_angles)  # True

    @pytest.mark.parametrize(
        ["alpha", "beta", "gamma"],
        [[126.96979, 54.83003, 178.20027], [44.82802, 110.52171, 65.693695]],
        ids=["over360", "close"],
    )
    def test_invalid_over360(self, alpha, beta, gamma):
        """Test for invalid unitcell angles, single frame"""
        unitcell_angles = np.array([alpha, beta, gamma], dtype=np.float32)

        assert not check_valid_unitcell_angles(*unitcell_angles)  # False

    def test_valid_array(self):
        """test for valid unitcell angles, input being an array"""
        unitcell_angles = np.array([[90.0, 90.0, 90], [70, 90, 30], [40, 20, 30]], dtype=np.float32)

        assert np.all(check_valid_unitcell_angles(*unitcell_angles.T) == [True, True, True])

    def test_invalid_over360_array(self):
        """Test for invalid unitcell angles that sum up to over 360˚, input being an array"""
        unitcell_angles = np.array([[90, 90, 90], [126.96979, 54.83003, 178.20027], [70, 90, 30]], dtype=np.float32)

        assert np.all(check_valid_unitcell_angles(*unitcell_angles.T) == [True, False, True])

    def test_invalid_angles_close_array(self):
        """Test for invalid unitcell angles where angles 1 and 3 sum up to more than angle 2"""
        unitcell_angles = np.array([[44.82802, 110.52171, 65.693695], [90, 90, 90], [70, 90, 30]], dtype=np.float64)

        assert np.all(check_valid_unitcell_angles(*unitcell_angles.T) == [False, True, True])



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2025 Stanford University and the Authors
#
# Authors: Robert McGibbon, Jeremy MG Leung
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import warnings

import numpy as np

##############################################################################
# Functions
##############################################################################


def lengths_and_angles_to_box_vectors(a_length, b_length, c_length, alpha, beta, gamma):
    """Convert from the lengths/angles of the unit cell to the box
    vectors (Bravais vectors). The angles should be in degrees.

    Parameters
    ----------
    a_length : scalar or np.ndarray
        length of Bravais unit vector **a**
    b_length : scalar or np.ndarray
        length of Bravais unit vector **b**
    c_length : scalar or np.ndarray
        length of Bravais unit vector **c**
    alpha : scalar or np.ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or np.ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or np.ndarray
        angle between vectors **a** and **b**, in degrees.

    Returns
    -------
    a : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    b : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    c : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)

    Examples
    --------
    >>> import numpy as np
    >>> result = lengths_and_angles_to_box_vectors(1, 1, 1, 90.0, 90.0, 90.0)

    Notes
    -----
    This code is adapted from gyroid, which is licensed under the BSD
    http://pythonhosted.org/gyroid/_modules/gyroid/unitcell.html
    """
    if np.all(alpha < 2 * np.pi) and np.all(beta < 2 * np.pi) and np.all(gamma < 2 * np.pi):
        warnings.warn("All your angles were less than 2*pi. Did you accidentally give me radians?")

    if not np.all(check_valid_unitcell_angles(alpha, beta, gamma)):
        warnings.warn("Certain frames have invalid unitcell box")

    alpha = np.radians(alpha)
    beta = np.radians(beta)
    gamma = np.radians(gamma)

    a = np.array([a_length, np.zeros_like(a_length), np.zeros_like(a_length)])
    b = np.array([b_length * np.cos(gamma), b_length * np.sin(gamma), np.zeros_like(b_length)])
    cx = c_length * np.cos(beta)
    cy = c_length * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma)
    cz = np.sqrt(c_length * c_length - cx * cx - cy * cy)
    c = np.array([cx, cy, cz])

    if not a.shape == b.shape == c.shape:
        raise TypeError("Shape is messed up.")

    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    a[np.logical_and(a > -tol, a < tol)] = 0.0
    b[np.logical_and(b > -tol, b < tol)] = 0.0
    c[np.logical_and(c > -tol, c < tol)] = 0.0

    return a.T, b.T, c.T


def box_vectors_to_lengths_and_angles(a, b, c):
    """Convert box vectors into the lengths and angles defining the box.

    Parameters
    ----------
    a : np.ndarray
        the vector defining the first edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where a[i,:] gives the
        length 3 array of vector a in each frame of a simulation
    b : np.ndarray
        the vector defining the second edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where b[i,:] gives the
        length 3 array of vector a in each frame of a simulation
    c : np.ndarray
        the vector defining the third edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where c[i,:] gives the
        length 3 array of vector a in each frame of a simulation

    Examples
    --------
    >>> a = np.array([2,0,0], dtype=float)
    >>> b = np.array([0,1,0], dtype=float)
    >>> c = np.array([0,1,1], dtype=float)
    >>> l1, l2, l3, alpha, beta, gamma = box_vectors_to_lengths_and_angles(a, b, c)
    >>> (l1 == 2.0) and (l2 == 1.0) and (l3 == np.sqrt(2))
    True
    >>> np.abs(alpha - 45) < 1e-6
    True
    >>> np.abs(beta - 90.0) < 1e-6
    True
    >>> np.abs(gamma - 90.0) < 1e-6
    True

    Returns
    -------
    a_length : scalar or np.ndarray
        length of Bravais unit vector **a**
    b_length : scalar or np.ndarray
        length of Bravais unit vector **b**
    c_length : scalar or np.ndarray
        length of Bravais unit vector **c**
    alpha : scalar or np.ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or np.ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or np.ndarray
        angle between vectors **a** and **b**, in degrees.
    """
    if not a.shape == b.shape == c.shape:
        raise TypeError("Shape is messed up.")
    if not a.shape[-1] == 3:
        raise TypeError("The last dimension must be length 3")
    if a.ndim not in [1, 2]:
        raise ValueError(
            "vectors must be 1d or 2d (for a vectorized operation on multiple frames)",
        )
    last_dim = a.ndim - 1

    a_length = np.sqrt(np.sum(a * a, axis=last_dim))
    b_length = np.sqrt(np.sum(b * b, axis=last_dim))
    c_length = np.sqrt(np.sum(c * c, axis=last_dim))

    # we allow 2d input, where the first dimension is the frame index
    # so we want to do the dot product only over the last dimension
    alpha = np.arccos(np.einsum("...i, ...i", b, c) / (b_length * c_length), casting="safe")
    beta = np.arccos(np.einsum("...i, ...i", c, a) / (c_length * a_length), casting="safe")
    gamma = np.arccos(np.einsum("...i, ...i", a, b) / (a_length * b_length), casting="safe")

    # convert to degrees
    alpha = np.degrees(alpha)
    beta = np.degrees(beta)
    gamma = np.degrees(gamma)

    return a_length, b_length, c_length, alpha, beta, gamma


def lengths_and_angles_to_tilt_factors(
    a_length,
    b_length,
    c_length,
    alpha,
    beta,
    gamma,
):
    """
    Parameters
    ----------
    a_length : scalar or np.ndarray
        length of Bravais unit vector **a**
    b_length : scalar or np.ndarray
        length of Bravais unit vector **b**
    c_length : scalar or np.ndarray
        length of Bravais unit vector **c**
    alpha : scalar or np.ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or np.ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or np.ndarray
        angle between vectors **a** and **b**, in degrees.

    Returns
    -------
    lx : scalar
        Extent in x direction
    ly : scalar
        Extent in y direction
    lz : scalar
        Extent in z direction
    xy : scalar
        Unit vector **b** tilt with respect to **a**
    xz : scalar
        Unit vector of **c** tilt with respect to **a**
    yz : scalar
        Unit vector of **c** tilt with respect to **b**
    """
    lx = a_length
    xy = b_length * np.cos(np.deg2rad(gamma))
    xz = c_length * np.cos(np.deg2rad(beta))
    ly = np.sqrt(b_length**2 - xy**2)
    yz = (b_length * c_length * np.cos(np.deg2rad(alpha)) - xy * xz) / ly
    lz = np.sqrt(c_length**2 - xz**2 - yz**2)

    return np.array([lx, ly, lz, xy, xz, yz])


def check_valid_unitcell_angles(alpha, beta, gamma):
    """Functions to check to see if unitcell is a valid triclinic simulation box.
    The unitcell angles are constrained to provide a positive volume,
    as shown in eq(4) of DOI:10.1107/S0108767310044296 or below::

        0 < alpha + beta + gamma < 360
        0 < sum(alpha, beta, gamma) - 2 * max(alpha, beta, gamma) < 360

    Parameters
    ----------
    alpha : scalar or np.ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or np.ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or np.ndarray
        angle between vectors **a** and **b**, in degrees.

    Returns
    -------
    results: np.bool_ or np.ndarray of np.bool_
        np.bool_ or an array of np.bool_ indicating whether input angles are valid
    """
    uca = np.dstack((alpha, beta, gamma))[0]
    uca_max = 2 * np.max(uca, axis=1)
    uca_sum = np.sum(uca, axis=1)

    results = np.ones(len(uca), dtype=bool)

    # Checking it row-by-row
    for row_idx, (row_max, row_sum) in enumerate(zip(uca_max, uca_sum)):
        if np.allclose(uca[row_idx], [90, 90, 90]):
            # Quick fast-track for rectilinear unitcells
            continue
        elif (
            not (0 < row_sum < 360)
            or not (row_max < row_sum)
            or not (row_sum < 360 + row_max)
            or np.isclose(0, row_sum)  # Check if angle sum's at the range edges
            or np.isclose(360, row_sum)
            or np.isclose(row_max, row_sum)
            or np.isclose(row_sum, row_max + 360)
        ):
            # Already defaults to True, so only need to modify to False
            results[row_idx] = False

    # Output based on whether input is float/int/scalar or list/array-like
    if isinstance(alpha, (np.floating, float, np.integer, int)):
        # Return just the True/False if input is scalar
        return results[0]
    else:
        # Return the whole array, assuming input is a list/array-like
        return results



"""
pdbxfile.py: Used for loading PDBx/mmCIF files.

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2015-2025 Stanford University and the Authors.
Authors: Peter Eastman
Contributors: Jason Swails

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

__author__ = "Peter Eastman"
__version__ = "2.0"

import sys
import warnings
from collections import defaultdict
from datetime import date

import numpy as np

import mdtraj._version
from mdtraj.core import element as elem
from mdtraj.core import topology
from mdtraj.core.topology import Topology
from mdtraj.utils import (
    open_maybe_zipped,
)

from .PdbxReader import PdbxReader


class PDBxFile:
    """PDBxFile parses a PDBx/mmCIF file and constructs a Topology and a set of atom positions from it."""

    def __init__(self, file):
        """Load a PDBx/mmCIF file.

        The atom positions and Topology can be retrieved by calling
        getPositions() and getTopology().

        Parameters
        ----------
        file : string
            the name of the file to load.  Alternatively you can pass an open
            file object.
        """
        top = Topology()
        ## The Topology read from the PDBx/mmCIF file
        self.topology = top
        self._positions = []
        self._unitcell_lengths = None
        self._unitcell_angles = None

        self._loadNameReplacementTables()

        # Load the file.

        inputFile = file
        ownHandle = False
        if isinstance(file, str):
            inputFile = open_maybe_zipped(file, "r")
            ownHandle = True
        reader = PdbxReader(inputFile)
        data = []
        reader.read(data)
        if ownHandle:
            inputFile.close()
        block = data[0]

        # Build the topology.

        atomData = block.getObj("atom_site")
        atomNameCol = atomData.getAttributeIndex("auth_atom_id")
        if atomNameCol == -1:
            atomNameCol = atomData.getAttributeIndex("label_atom_id")
        atomIdCol = atomData.getAttributeIndex("id")
        resNameCol = atomData.getAttributeIndex("auth_comp_id")
        if resNameCol == -1:
            resNameCol = atomData.getAttributeIndex("label_comp_id")
        resIdCol = atomData.getAttributeIndex("label_seq_id")
        resNumCol = atomData.getAttributeIndex("auth_seq_id")
        if resNumCol == -1:
            resNumCol = resIdCol
        resInsertionCol = atomData.getAttributeIndex("pdbx_PDB_ins_code")
        chainIdCol = atomData.getAttributeIndex("auth_asym_id")
        if chainIdCol == -1:
            chainIdCol = atomData.getAttributeIndex("label_asym_id")
            altChainIdCol = -1
        else:
            altChainIdCol = atomData.getAttributeIndex("label_asym_id")
        if altChainIdCol != -1:
            # Figure out which column is best to use for chain IDs.

            idSet = set(row[chainIdCol] for row in atomData.getRowList())
            altIdSet = set(row[altChainIdCol] for row in atomData.getRowList())
            if len(altIdSet) > len(idSet):
                chainIdCol, altChainIdCol = (altChainIdCol, chainIdCol)
        elementCol = atomData.getAttributeIndex("type_symbol")
        altIdCol = atomData.getAttributeIndex("label_alt_id")
        modelCol = atomData.getAttributeIndex("pdbx_PDB_model_num")
        xCol = atomData.getAttributeIndex("Cartn_x")
        yCol = atomData.getAttributeIndex("Cartn_y")
        zCol = atomData.getAttributeIndex("Cartn_z")
        lastChainId = None
        lastAltChainId = None
        lastResId = None
        lastInsertionCode = ""
        atomTable = {}
        atomsInResidue = set()
        models = []

        for row in atomData.getRowList():
            atomKey = (row[resNumCol], row[chainIdCol], row[atomNameCol])
            model = "1" if modelCol == -1 else row[modelCol]
            if model not in models:
                models.append(model)
                self._positions.append([])
            modelIndex = models.index(model)
            if (
                row[altIdCol] != "."
                and atomKey in atomTable
                and len(self._positions[modelIndex]) > atomTable[atomKey].index
            ):
                # This row is an alternate position for an existing atom, so ignore it.

                continue
            if modelIndex == 0:
                # This row defines a new atom.

                if resInsertionCol == -1:
                    insertionCode = ""
                else:
                    insertionCode = row[resInsertionCol]
                if insertionCode in (".", "?"):
                    insertionCode = ""
                if lastChainId != row[chainIdCol] or (altChainIdCol != -1 and lastAltChainId != row[altChainIdCol]):
                    # The start of a new chain.
                    chain = top.add_chain(row[chainIdCol])
                    lastChainId = row[chainIdCol]
                    lastResId = None
                    if altChainIdCol != -1:
                        lastAltChainId = row[altChainIdCol]
                if (
                    lastResId != row[resNumCol]
                    or lastChainId != row[chainIdCol]
                    or lastInsertionCode != insertionCode
                    or (lastResId == "." and row[atomNameCol] in atomsInResidue)
                ):
                    # The start of a new residue.
                    resId = None if resNumCol == -1 else int(row[resNumCol])
                    resIC = insertionCode
                    resName = row[resNameCol]
                    if resName in self._residueNameReplacements:
                        resName = self._residueNameReplacements[resName]
                    res = top.add_residue(resName, chain, resId, resIC)
                    if resName in self._atomNameReplacements:
                        atomReplacements = self._atomNameReplacements[resName]
                    else:
                        atomReplacements = {}
                    lastResId = row[resNumCol]
                    lastInsertionCode = insertionCode
                    atomsInResidue.clear()
                element = None
                try:
                    element = elem.get_by_symbol(row[elementCol])
                except KeyError:
                    pass
                atomName = row[atomNameCol]
                if atomName in atomReplacements:
                    atomName = atomReplacements[atomName]
                atom = top.add_atom(atomName, element, res, row[atomIdCol])
                atomTable[atomKey] = atom
                atomsInResidue.add(atomName)
            else:
                # This row defines coordinates for an existing atom in one of the later models.
                try:
                    atom = atomTable[atomKey]
                except KeyError:
                    raise ValueError(
                        f"Unknown atom {row[atomNameCol]} in residue "
                        f"{row[resNameCol]} {row[resNumCol]} for model {model}",
                    )
                if atom.index != len(self._positions[modelIndex]):
                    raise ValueError(
                        f"Atom {row[atomIdCol]} for model {model} "
                        f"does not match the order of atoms for model {models[0]}",
                    )
            pos = np.array([float(row[xCol]), float(row[yCol]), float(row[zCol])]) / 10  # nanometers
            self._positions[modelIndex].append(pos)
        ## The atom positions read from the PDBx/mmCIF file.
        ## If the file contains multiple frames, these are the positions in the first frame.
        self.positions = self._positions[0]
        self._numpyPositions = None

        # Record unit cell information, if present.
        cell = block.getObj("cell")
        if cell is not None and cell.getRowCount() > 0:
            row = cell.getRow(0)
            self._unitcell_lengths = [
                float(row[cell.getAttributeIndex(attribute)]) / 10
                for attribute in ("length_a", "length_b", "length_c")
            ]  # nanometers
            self._unitcell_angles = [
                float(row[cell.getAttributeIndex(attribute)])
                for attribute in ("angle_alpha", "angle_beta", "angle_gamma")
            ]  # degrees

        # Add bonds for standard residues.
        self.topology.create_standard_bonds()

        # Add bonds based on struct_conn records.
        connectData = block.getObj("struct_conn")
        if connectData is not None:
            res1Col = connectData.getAttributeIndex("ptnr1_label_seq_id")
            res2Col = connectData.getAttributeIndex("ptnr2_label_seq_id")
            atom1Col = connectData.getAttributeIndex("ptnr1_label_atom_id")
            atom2Col = connectData.getAttributeIndex("ptnr2_label_atom_id")
            asym1Col = connectData.getAttributeIndex("ptnr1_label_asym_id")
            asym2Col = connectData.getAttributeIndex("ptnr2_label_asym_id")
            typeCol = connectData.getAttributeIndex("conn_type_id")
            connectBonds = []
            for row in connectData.getRowList():
                type = row[typeCol][:6]
                if type in ("covale", "disulf", "modres"):
                    key1 = (row[res1Col], row[asym1Col], row[atom1Col])
                    key2 = (row[res2Col], row[asym2Col], row[atom2Col])
                    if key1 in atomTable and key2 in atomTable:
                        connectBonds.append((atomTable[key1], atomTable[key2]))
            if len(connectBonds) > 0:
                # Only add bonds that don't already exist.
                existingBonds = set(top.bonds)
                for bond in connectBonds:
                    if bond not in existingBonds and (bond[1], bond[0]) not in existingBonds:
                        top.add_bond(bond[0], bond[1])
                        existingBonds.add(bond)

        # Add bonds based on chem_comp_bond records.
        bondData = block.getObj("chem_comp_bond")
        if bondData is not None:
            # Load the bond definitions for residues.

            resNameCol = bondData.getAttributeIndex("comp_id")
            atom1Col = bondData.getAttributeIndex("atom_id_1")
            atom2Col = bondData.getAttributeIndex("atom_id_2")
            bondOrderCol = bondData.getAttributeIndex("value_order")
            resBonds = defaultdict(list)
            for row in bondData.getRowList():
                bondOrder = None if bondOrderCol == -1 else row[bondOrderCol]
                resBonds[row[resNameCol]].append(
                    (row[atom1Col], row[atom2Col], bondOrder),
                )

            # Create the bonds.

            bondTypes = defaultdict(
                lambda: None,
                {
                    "sing": topology.Single,
                    "doub": topology.Double,
                    "trip": topology.Triple,
                    "arom": topology.Aromatic,
                },
            )
            bondOrders = defaultdict(lambda: None, {"sing": 1, "doub": 2, "trip": 3})
            for res in self.topology.residues:
                if res.name in resBonds:
                    atoms = {atom.name: atom for atom in res.atoms}
                    for atom1, atom2, bondOrder in resBonds[res.name]:
                        if atom1 in atoms and atom2 in atoms:
                            self.topology.add_bond(
                                atoms[atom1],
                                atoms[atom2],
                                bondTypes[bondOrder],
                                bondOrders[bondOrder],
                            )

        # Remove any duplicate bonds
        bonds = self.topology._bonds

        self.topology._bonds = []
        unique_bonds = {}
        for bond in bonds:
            i = bond.atom1.index
            j = bond.atom2.index
            if j < i:
                i, j = j, i
            if (i, j) not in unique_bonds:
                unique_bonds[(i, j)] = bond
            else:
                # If the bond already exists use the one containing the most information or the last one.
                if unique_bonds[(i, j)].order is None:
                    unique_bonds[(i, j)].order = bond.order
                elif bond.order is not None and unique_bonds[(i, j)].order != bond.order:
                    warnings.warn(
                        f"Bond {bond.atom1}-{bond.atom2} order ({unique_bonds[(i, j)].order} "
                        f"will be replaced by {bond.order})",
                    )
                    unique_bonds[(i, j)].order = bond.order

                if unique_bonds[(i, j)].type is None:
                    unique_bonds[(i, j)].type = bond.type
                elif bond.type is not None and unique_bonds[(i, j)].type != bond.type:
                    warnings.warn(
                        f"Bond {bond.atom1}-{bond.atom2} type ({unique_bonds[(i, j)].type} "
                        f"will be replaced by {bond.type})",
                    )
                    unique_bonds[(i, j)].type = bond.type

        self.topology._bonds = list(unique_bonds.values())

    def getTopology(self):
        """Get the Topology of the model."""
        return self.topology

    def getNumFrames(self):
        """Get the number of frames stored in the file."""
        return len(self._positions)

    def getPositions(self, asNumpy=False, frame=0):
        """Get the atomic positions

        Parameters
        ----------
        asNumpy : bool=False
            if true, the values are returned as a numpy array instead of a list
            of Vec3s
        frame : int=0
            the index of the frame for which to get positions
        """
        if asNumpy:
            if self._numpyPositions is None:
                self._numpyPositions = [None] * len(self._positions)
            if self._numpyPositions[frame] is None:
                self._numpyPositions[frame] = self._positions[frame]
            return self._numpyPositions[frame]
        return self._positions[frame]

    @staticmethod
    def writeFile(
        topology,
        positions,
        box_lengths=None,
        box_angles=None,
        file=sys.stdout,
        keepIds=False,
        entry=None,
        bfactors=None,
    ):
        """Write a PDBx/mmCIF file containing a single model.

        Parameters
        ----------
        topology : Topology
            The Topology defining the model to write
        positions : list
            The list of atomic positions to write
        box_lengths : list=None
            The lengths of the periodic box vectors in nanometers
        box_angles : list=None
            The angles of the periodic box vectors in radians
        file : string or file
            the name of the file to write.  Alternatively you can pass an open file object.
        keepIds : bool=False
            If True, keep the residue and chain IDs specified in the Topology
            rather than generating new ones.  Warning: It is up to the caller to
            make sure these are valid IDs that satisfy the requirements of the
            PDBx/mmCIF format.  Otherwise, the output file will be invalid.
        entry : str=None
            The entry ID to assign to the CIF file
        bfactors : list=None
            The list of B-factors to write for each atom. If None, all B-factors
            will be set to 0.0.
        """
        if isinstance(file, str):
            with open(file, "w") as output:
                PDBxFile.writeFile(topology, positions, output, keepIds, entry)
        else:
            PDBxFile.writeHeader(
                topology,
                box_lengths,
                box_angles,
                file,
                entry,
                keepIds,
            )
            PDBxFile.writeModel(topology, positions, file, keepIds=keepIds, bfactors=bfactors)

    @staticmethod
    def writeHeader(
        topology,
        box_lengths=None,
        box_angles=None,
        file=sys.stdout,
        entry=None,
        keepIds=False,
    ):
        """Write out the header for a PDBx/mmCIF file.

        Parameters
        ----------
        topology : Topology
            The Topology defining the molecular system being written
        box_lengths : list=None
            The lengths of the periodic box vectors in nanometers
        box_angles : list=None
            The angles of the periodic box vectors in degrees
        file : file=stdout
            A file to write the file to
        entry : str=None
            The entry ID to assign to the CIF file
        keepIds : bool=False
            If True, keep the residue and chain IDs specified in the Topology
            rather than generating new ones.  Warning: It is up to the caller to
            make sure these are valid IDs that satisfy the requirements of the
            PDBx/mmCIF format.  Otherwise, the output file will be invalid.
        """
        if entry is not None:
            print(f"data_{entry}", file=file)
        else:
            print("data_cell", file=file)
        print(
            f"# Created with mdtraj {mdtraj._version.get_versions()['version']}, {str(date.today())}",
            file=file,
        )
        print("#", file=file)

        if box_lengths is not None and box_angles is not None:
            a, b, c = box_lengths
            alpha, beta, gamma = box_angles
            print(f"_cell.length_a     {a * 10:10.4f}", file=file)
            print(f"_cell.length_b     {b * 10:10.4f}", file=file)
            print(f"_cell.length_c     {c * 10:10.4f}", file=file)
            print(f"_cell.angle_alpha  {alpha:10.4f}", file=file)
            print(f"_cell.angle_beta   {beta:10.4f}", file=file)
            print(f"_cell.angle_gamma  {gamma:10.4f}", file=file)
            print("#", file=file)

        # Identify bonds that should be listed in the file.

        bonds = []
        for atom1, atom2 in topology.bonds:
            if (
                atom1.residue.name not in PDBxFile._standardResidues
                or atom2.residue.name not in PDBxFile._standardResidues
            ):
                bonds.append((atom1, atom2))
            elif (
                atom1.name == "SG"
                and atom2.name == "SG"
                and atom1.residue.name == "CYS"
                and atom2.residue.name == "CYS"
            ):
                bonds.append((atom1, atom2))
        if len(bonds) > 0:
            # Write the bond information.

            print("loop_", file=file)
            print("_struct_conn.id", file=file)
            print("_struct_conn.conn_type_id", file=file)
            print("_struct_conn.ptnr1_label_asym_id", file=file)
            print("_struct_conn.ptnr1_label_comp_id", file=file)
            print("_struct_conn.ptnr1_label_seq_id", file=file)
            print("_struct_conn.ptnr1_label_atom_id", file=file)
            print("_struct_conn.ptnr2_label_asym_id", file=file)
            print("_struct_conn.ptnr2_label_comp_id", file=file)
            print("_struct_conn.ptnr2_label_seq_id", file=file)
            print("_struct_conn.ptnr2_label_atom_id", file=file)
            chainIds = {}
            resIds = {}
            if keepIds:
                for chain in topology.chains:
                    chainIds[chain.index] = chain.chain_id
                for res in topology.residues:
                    # or alt could set to res.index + 1
                    resIds[res.index] = res.resSeq
            else:
                for chainIndex, chain in enumerate(topology.chains):
                    chainIds[chain.index] = chr(ord("A") + chainIndex % 26)
                    for resIndex, res in enumerate(chain.residues):
                        resIds[res.index] = resIndex + 1
            for i, (atom1, atom2) in enumerate(bonds):
                if atom1.residue.name == "CYS" and atom2.residue.name == "CYS":
                    bondType = "disulf"
                else:
                    bondType = "covale"
                line = "bond%d %s %s %-4s %5s %-4s %s %-4s %5s %-4s"
                print(
                    line
                    % (
                        i + 1,
                        bondType,
                        chainIds[atom1.residue.chain.index],
                        atom1.residue.name,
                        resIds[atom1.residue.index],
                        atom1.name,
                        chainIds[atom2.residue.chain.index],
                        atom2.residue.name,
                        resIds[atom2.residue.index],
                        atom2.name,
                    ),
                    file=file,
                )
            print("#", file=file)

        # Write the header for the atom coordinates.

        print("loop_", file=file)
        print("_atom_site.group_PDB", file=file)
        print("_atom_site.id", file=file)
        print("_atom_site.type_symbol", file=file)
        print("_atom_site.label_atom_id", file=file)
        print("_atom_site.label_alt_id", file=file)
        print("_atom_site.label_comp_id", file=file)
        print("_atom_site.label_asym_id", file=file)
        print("_atom_site.label_entity_id", file=file)
        print("_atom_site.label_seq_id", file=file)
        print("_atom_site.pdbx_PDB_ins_code", file=file)
        print("_atom_site.Cartn_x", file=file)
        print("_atom_site.Cartn_y", file=file)
        print("_atom_site.Cartn_z", file=file)
        print("_atom_site.occupancy", file=file)
        print("_atom_site.B_iso_or_equiv", file=file)
        print("_atom_site.Cartn_x_esd", file=file)
        print("_atom_site.Cartn_y_esd", file=file)
        print("_atom_site.Cartn_z_esd", file=file)
        print("_atom_site.occupancy_esd", file=file)
        print("_atom_site.B_iso_or_equiv_esd", file=file)
        print("_atom_site.pdbx_formal_charge", file=file)
        print("_atom_site.auth_seq_id", file=file)
        print("_atom_site.auth_comp_id", file=file)
        print("_atom_site.auth_asym_id", file=file)
        print("_atom_site.auth_atom_id", file=file)
        print("_atom_site.pdbx_PDB_model_num", file=file)

    @staticmethod
    def writeModel(topology, positions, file=sys.stdout, modelIndex=1, keepIds=False, bfactors=None):
        """Write out a model to a PDBx/mmCIF file.

        Parameters
        ----------
        topology : Topology
            The Topology defining the model to write
        positions : list
            The list of atomic positions to write
        file : file=stdout
            A file to write the model to
        modelIndex : int=1
            The model number of this frame
        keepIds : bool=False
            If True, keep the residue and chain IDs specified in the Topology
            rather than generating new ones.  Warning: It is up to the caller to
            make sure these are valid IDs that satisfy the requirements of the
            PDBx/mmCIF format.  Otherwise, the output file will be invalid.
        bfactors : list=None
            The list of B-factors to write for each atom. If None, all B-factors
            will be set to 0.0.
        """
        cif_line = (
            "%s  %5d %-3s %-4s . %-4s %s ? %5s %s %10.4f %10.4f %10.4f 0.0 %5.2f  ?  ?  ?  ?  ?  .  %5s %4s %s %4s %5d"
        )

        if len(list(topology.atoms)) != len(positions):
            raise ValueError("The number of positions must match the number of atoms")
        positions = np.asarray(positions) * 10  # convert to angstroms
        if np.isnan(positions).any():
            raise ValueError(
                "Particle position is NaN.  For more information, see https://github.com/openmm/openmm/wiki/Frequently-Asked-Questions#nan",
            )
        if np.isinf(positions).any():
            raise ValueError(
                "Particle position is infinite.  For more information, see https://github.com/openmm/openmm/wiki/Frequently-Asked-Questions#nan",
            )
        nonHeterogens = PDBxFile._standardResidues[:]
        nonHeterogens.remove("HOH")
        atomIndex = 1
        posIndex = 0

        if bfactors is None:
            bfactors = [0.0] * len(positions)

        for chainIndex, chain in enumerate(topology.chains):
            if keepIds:
                chainName = chain.chain_id
            else:
                chainName = chr(ord("A") + chainIndex % 26)
            residues = list(chain.residues)
            for resIndex, res in enumerate(residues):
                if keepIds:
                    resId = res.resSeq
                    resIC = res.insertionCode.strip() if hasattr(res, "insertionCode") else "."
                else:
                    resId = resIndex + 1
                    resIC = "."
                if res.name in nonHeterogens:
                    recordName = "ATOM"
                else:
                    recordName = "HETATM"
                for atom in res.atoms:
                    coords = positions[posIndex]
                    if atom.element is not None:
                        symbol = atom.element.symbol
                    else:
                        symbol = "?"
                    print(
                        cif_line
                        % (
                            recordName,
                            atomIndex,
                            symbol,
                            atom.name,
                            res.name,
                            chainName,
                            resId,
                            resIC,
                            coords[0],
                            coords[1],
                            coords[2],
                            bfactors[posIndex],
                            resId,
                            res.name,
                            chainName,
                            atom.name,
                            modelIndex,
                        ),
                        file=file,
                    )
                    posIndex += 1
                    atomIndex += 1
        print("#", file=file)

    ## TODO: Methods repeated in PDBTrajectoryFile. They should be moved to a common place.
    _standardResidues = [
        "ALA",
        "ASN",
        "CYS",
        "GLU",
        "HIS",
        "LEU",
        "MET",
        "PRO",
        "THR",
        "TYR",
        "ARG",
        "ASP",
        "GLN",
        "GLY",
        "ILE",
        "LYS",
        "PHE",
        "SER",
        "TRP",
        "VAL",
        "A",
        "G",
        "C",
        "U",
        "I",
        "DA",
        "DG",
        "DC",
        "DT",
        "DI",
        "HOH",
    ]
    _residueNameReplacements = {}
    _atomNameReplacements = {}

    @staticmethod
    def _loadNameReplacementTables():
        import os
        import xml.etree.ElementTree as etree
        from copy import copy

        """Load the list of atom and residue name replacements."""
        if len(PDBxFile._residueNameReplacements) == 0:
            tree = etree.parse(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "pdb",
                    "data",
                    "pdbNames.xml",
                ),
            )
            allResidues = {}
            proteinResidues = {}
            nucleicAcidResidues = {}
            for residue in tree.getroot().findall("Residue"):
                name = residue.attrib["name"]
                if name == "All":
                    PDBxFile._parseResidueAtoms(residue, allResidues)
                elif name == "Protein":
                    PDBxFile._parseResidueAtoms(residue, proteinResidues)
                elif name == "Nucleic":
                    PDBxFile._parseResidueAtoms(residue, nucleicAcidResidues)
            for atom in allResidues:
                proteinResidues[atom] = allResidues[atom]
                nucleicAcidResidues[atom] = allResidues[atom]
            for residue in tree.getroot().findall("Residue"):
                name = residue.attrib["name"]
                for id in residue.attrib:
                    if id == "name" or id.startswith("alt"):
                        PDBxFile._residueNameReplacements[residue.attrib[id]] = name
                if "type" not in residue.attrib:
                    atoms = copy(allResidues)
                elif residue.attrib["type"] == "Protein":
                    atoms = copy(proteinResidues)
                elif residue.attrib["type"] == "Nucleic":
                    atoms = copy(nucleicAcidResidues)
                else:
                    atoms = copy(allResidues)
                PDBxFile._parseResidueAtoms(residue, atoms)
                PDBxFile._atomNameReplacements[name] = atoms

    @staticmethod
    def _parseResidueAtoms(residue, map):
        for atom in residue.findall("Atom"):
            name = atom.attrib["name"]
            for id in atom.attrib:
                map[atom.attrib[id]] = name



import os
import sys

import mdtraj


def test_1():
    capi = mdtraj.capi()
    assert os.path.isdir(capi["lib_dir"])
    assert os.path.isdir(capi["include_dir"])
    if sys.platform == "win32":
        assert os.path.isfile(os.path.join(capi["lib_dir"], "libtheobald.lib"))
    else:
        assert os.path.isfile(os.path.join(capi["lib_dir"], "libtheobald.a"))



#!/bin/env python
"""
Module openmm.unit.math

Arithmetic methods on Quantities and Units

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012 Stanford University and the Authors.
Authors: Christopher M. Bruns
Contributors: Peter Eastman

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

__author__ = "Christopher M. Bruns"
__version__ = "0.5"


import math

from .quantity import is_quantity
from .unit_definitions import *

####################
### TRIGONOMETRY ###
####################


def sin(angle):
    """
    Examples

    >>> sin(90*degrees)
    1.0
    """
    if is_quantity(angle):
        return math.sin(angle / radians)
    else:
        return math.sin(angle)


def sinh(angle):
    if is_quantity(angle):
        return math.sinh(angle / radians)
    else:
        return math.sinh(angle)


def cos(angle):
    """
    Examples

    >>> cos(180*degrees)
    -1.0
    """
    if is_quantity(angle):
        return math.cos(angle / radians)
    else:
        return math.cos(angle)


def cosh(angle):
    if is_quantity(angle):
        return math.cosh(angle / radians)
    else:
        return math.cosh(angle)


def tan(angle):
    if is_quantity(angle):
        return math.tan(angle / radians)
    else:
        return math.tan(angle)


def tanh(angle):
    if is_quantity(angle):
        return math.tanh(angle / radians)
    else:
        return math.tanh(angle)


def acos(x):
    """
    >>> acos(1.0)
    Quantity(value=0.0, unit=radian)
    >>> print(acos(1.0))
    0.0 rad
    """
    return math.acos(x) * radians


def acosh(x):
    return math.acosh(x) * radians


def asin(x):
    return math.asin(x) * radians


def asinh(x):
    return math.asinh(x) * radians


def atan(x):
    return math.atan(x) * radians


def atanh(x):
    return math.atanh(x) * radians


def atan2(x, y):
    return math.atan2(x, y) * radians


###################
### SQUARE ROOT ###
###################


def sqrt(val):
    """
    >>> sqrt(9.0)
    3.0
    >>> print(sqrt(meter*meter))
    meter
    >>> sqrt(9.0*meter*meter)
    Quantity(value=3.0, unit=meter)
    >>> sqrt(9.0*meter*meter*meter)
    Traceback (most recent call last):
    ...
    ArithmeticError: Exponents in Unit.sqrt() must be even.
    """
    try:
        return val.sqrt()
    except AttributeError:
        return math.sqrt(val)


###########
### SUM ###
###########


def sum(val):
    """
    >>> sum((1.0, 2.0))
    3.0
    >>> sum((2.0*meter, 3.0*meter))
    Quantity(value=5.0, unit=meter)
    >>> sum((2.0*meter, 30.0*centimeter))
    Quantity(value=2.3, unit=meter)
    """
    try:
        return val.sum()
    except AttributeError:
        pass
    if len(val) == 0:
        return 0
    result = val[0]
    for i in range(1, len(val)):
        result += val[i]
    return result


###################
### VECTOR MATH ###
###################


def dot(x, y):
    """
    >>> dot((2, 3)*meter, (4, 5)*meter)
    Quantity(value=23, unit=meter**2)
    """
    sum = x[0] * y[0]
    for i in range(1, len(x)):
        sum += x[i] * y[i]
    return sum


def norm(x):
    """
    >>> norm((3, 4)*meter)
    Quantity(value=5.0, unit=meter)
    """
    return sqrt(dot(x, x))


# run module directly for testing
if __name__ == "__main__":
    # Test the examples in the docstrings
    import doctest
    import sys

    doctest.testmod(sys.modules[__name__])



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2025 Stanford University and the Authors
#
# Authors: Lee-Ping Wang
# Contributors: Robert McGibbon, Jeremy M. G. Leung
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import numpy as np

import mdtraj as md
from mdtraj.formats import ArcTrajectoryFile, PDBTrajectoryFile
from mdtraj.formats.arc import load_arc
from mdtraj.testing import eq


def test_read_0(get_fn):
    """Check if .arc positions are read correctly"""
    with ArcTrajectoryFile(get_fn("4waters.arc")) as f:
        xyz, leng, ang = f.read()
    with PDBTrajectoryFile(get_fn("4waters.pdb")) as f:
        xyz2 = f.positions
    eq(xyz, xyz2, decimal=3)


def test_read_arctraj(get_fn):
    traj = md.load(get_fn("nitrogen.arc"), top=get_fn("nitrogen.pdb"))
    owntop = md.load(get_fn("nitrogen.arc"))
    eq(traj.xyz, owntop.xyz)


def test_read_arctraj_top(get_fn):
    """Check if anything other than residues differ between loading
    a ``.arc`` file with two different topologies"""
    pdbtop = load_arc(get_fn("4waters.arc"), top=get_fn("4waters.pdb"))
    arctop = load_arc(get_fn("4waters.arc"))

    eq(pdbtop.n_atoms, arctop.n_atoms)
    eq(pdbtop.xyz, arctop.xyz)

    # Only the residues should differ
    assert pdbtop.n_residues == 4
    assert arctop.n_residues == 1

    for ipdb, iarc in zip(pdbtop.top._atoms, arctop.top._atoms):
        # check if atoms are in the same order with same element identity
        eq(ipdb.name, iarc.name)
        eq(ipdb.element, iarc.element)

    for ipdb, iarc in zip(pdbtop.top._bonds, arctop.top._bonds):
        # Check if connectivity is the same, based on index, which is
        # implicitly checked in the for loop above
        eq(ipdb.atom1.index, iarc.atom1.index)
        eq(ipdb.atom2.index, iarc.atom2.index)


def test_read_pbc(get_fn):
    with ArcTrajectoryFile(get_fn("thf200.arc")) as f:
        xyz, leng, ang = f.read()
    eq(xyz[0, 0], np.array([21.231166, 32.549819, 8.278454]), decimal=3)
    eq(xyz[0, -1], np.array([29.013880, 6.255754, 44.074519]), decimal=3)
    eq(leng, np.array([[45.119376, 45.119376, 45.119376]]), decimal=3)
    eq(ang, np.array([[90.0, 90.0, 90.0]]), decimal=3)
